@@ -17,7 +17,7 @@
   const { BadResource, Interrupted } = core;
   const { ReadableStream } = window.__bootstrap.streams;
   const abortSignal = window.__bootstrap.abortSignal;
-  const { WebSocket, _rid, _readyState, _eventLoop, _protocol } =
+  const { WebSocket, _rid, _readyState, _eventLoop, _protocol, _server } =
     window.__bootstrap.webSocket;
   const {
     ArrayPrototypeIncludes,
@@ -368,11 +368,12 @@
 
     const response = fromInnerResponse(r, "immutable");
 
-    const websocket = webidl.createBranded(WebSocket);
-    setEventTargetData(websocket);
-    response[_ws] = websocket;
+    const socket = webidl.createBranded(WebSocket);
+    setEventTargetData(socket);
+    socket[_server] = true;
+    response[_ws] = socket;
 
-    return { response, websocket };
+    return { response, socket };
   }
 
   window.__bootstrap.http = {
